#include "SYSCALLS.H"
#include <Windows.h>

HANDLE hConsole;

int int86(int intno, union Registers *inregs, union Registers *outregs) {
  DWORD cCharsWritten;
  WORD attribute;
  CONSOLE_SCREEN_BUFFER_INFO csbiInfo;
  COORD pos;
  if (intno == 0x10) {
    switch (inregs->h.ah) {
      case 0x0A:
        /* Int 10/AH=0Ah: VIDEO - WRITE CHARACTER ONLY AT CURSOR POSITION
           al = character to write
           bh = page 0
           cx = number of times to write */
        GetConsoleScreenBufferInfo(hConsole, &csbiInfo);
        FillConsoleOutputCharacterA(hConsole,  // Handle to console screen buffer
          (TCHAR)inregs->h.al,                 // Character to write to the buffer
          inregs->x.cx,                        // Number of chars to write
          csbiInfo.dwCursorPosition,           // Coordinates of first chars
          &cCharsWritten);
        break;
      case 0x02:
        /* Int 10/AH=02h: VIDEO - SET CURSOR POSITION
           bh = page 0
           dh = row (00h is top)
           dl = column (00h is left) */
        pos.X = inregs->h.dl;
        pos.Y = inregs->h.dh;
        SetConsoleCursorPosition(hConsole, pos);
        break;
      case 0x0E:
        /* Int 10/AH=0Eh: VIDEO - TELETYPE OUTPUT
           Display a character on the screen, advancing the cursor and scrolling the screen as necessary  
           al = character to write
           bh = page 0 */
        GetConsoleScreenBufferInfo(hConsole, &csbiInfo);
        FillConsoleOutputCharacterA(hConsole,  // Handle to console screen buffer
          (TCHAR)inregs->h.al,                 // Character to write to the buffer
          1,                                   // Number of chars to write
          csbiInfo.dwCursorPosition,           // Coordinates of first chars
          &cCharsWritten);
        csbiInfo.dwCursorPosition.X++;         /* TODO: This should handle wrapping to the next line, but we don't need it now */
        SetConsoleCursorPosition(hConsole, csbiInfo.dwCursorPosition);
        break;
      case 0x0F:
        /* Int 10/AH=0Fh: VIDEO - GET CURRENT VIDEO MODE 
		   al = display mode
		   ah = number of columns
		   bh = active page */
        outregs->h.al = 3;
        outregs->h.bh = 0;
		/* TODO: This should put number of columns in ah, but we don't need it now */
        break;
      case 0x06:
        /* Int 10/AH=06h: VIDEO - SCROLL UP WINDOW 
           al = number of lines (0: clear whole window), 
           ch/cl = row/column upper left, 
           dh/dl = row/column lower right, 
           bh = attribute to write */
        if (inregs->h.al == 0) {
          pos.X = inregs->h.cl;
          pos.Y = inregs->h.ch;
          GetConsoleScreenBufferInfo(hConsole, &csbiInfo);
          if (inregs->h.dl == 80 && inregs->h.dh == 25) {
            /* Detected 80x25 so adjust to actual console size */
            inregs->h.dl = csbiInfo.dwSize.X;
            inregs->h.dh = csbiInfo.dwSize.Y;
          }
          FillConsoleOutputCharacterA(hConsole,                            // Handle to console screen buffer
            ' ',                                                           // Character to write to the buffer
            (inregs->h.dl - inregs->h.cl) * (inregs->h.dh - inregs->h.ch), // Number of chars to write
            pos,                                                           // Coordinates of first chars
            &cCharsWritten);
          FillConsoleOutputAttribute(hConsole,                             // Handle to console screen buffer
            inregs->h.bh,                                                  // Attribute to write to the buffer
            (inregs->h.dl - inregs->h.cl) * (inregs->h.dh - inregs->h.ch), // Number of chars to write
            pos,                                                           // Coordinates of first chars
            &cCharsWritten);
        }
		/* TODO: This should handle cases other than clearing the whole window, but we don't need it now */
        break;
      case 0x08:
        /* Int 10/AH=08h: VIDEO - READ CHARACTER AND ATTRIBUTE AT CURSOR POSITION
           bh = page 0
           ah = character's attribute (output)
           al = character (output) */
        GetConsoleScreenBufferInfo(hConsole, &csbiInfo);
        outregs->h.ah = csbiInfo.wAttributes; // Attribute
		/* TODO: This should place character in al, but we don't need it now */
        break;
    }
  } else if (intno == 0x15) {
    /* Int 15/AH=4Dh: HP-95LX detection, not supported */
  }
  return 0;
}

void _fmemmove2(int *dest, int *source, size_t count) {
  unsigned int i, character, color;
  WORD attribute;
  COORD pos;
  DWORD cCharsWritten;
  const WORD FOREGROUND_WHITE = FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE;
  const WORD FOREGROUND_YELLOW = FOREGROUND_RED | FOREGROUND_GREEN;

  for (i  = 0; i < count; i++) {
    character = source[i] & 0xFF;
    color = (source[i] & 0xFF00) >> 8;
    switch (color) {
        case 0x1C: attribute = FOREGROUND_RED; break;
        case 0x1E: attribute = FOREGROUND_YELLOW; break;
        case 0x1F: attribute = FOREGROUND_WHITE; break;
        default: attribute = 0; break;
    }
    pos.X = i%80;
    pos.Y = i/80;
    attribute = attribute | BACKGROUND_BLUE | FOREGROUND_INTENSITY;
    FillConsoleOutputCharacterA(hConsole,  // Handle to console screen buffer
      (TCHAR)character,                    // Character to write to the buffer
      1,                                   // Number of chars to write
      pos,                                 // Coordinates of first chars
      &cCharsWritten);
    FillConsoleOutputAttribute(hConsole,   // Handle to console screen buffer
      attribute,                           // Attribute to write to the buffer
      1,                                   // Number of chars to write
      pos,                                 // Coordinates of first chars
      &cCharsWritten);
  }
}